# Pico SDK

The Pico SDK provides the headers, libraries and build system
necessary to write programs for the RP2040 based devices such as the Raspberry Pi Pico
in C, C++ or assembly language.

The Pico SDK  is designed to provide an API and programming environment that is familiar both to non-embedded C developers and embedded C developers alike.
A single program runs on the device at a time with a conventional `main()` method. Standard C/C++ libraries are supported along with
APIs for accessing the RP2040's hardware, including DMA, IRQs, and the wide variety fixed function peripherals and PIO (Programmable IO).

Additionally the Pico SDK provides higher level libraries for dealing with timers, USB, synchronization and multi-core programming, along with
additional high level functionality built using PIO such as audio.

The Pico SDK can be used to build anything from simple applications, full fledged runtime environments such as MicroPython, to low level software
such as the RP2040's on chip bootrom itself.

# Documentation

See [Getting Started with the Raspberry Pi Pico](rptl.io/picogs) for information on how to setup your
hardware, IDE/environment and for how to build and debug software for the Raspberry Pi Pico
and other RP2040 based devices.

See [Pico C/C++ SDK](https://rptl.io/picoc) to learn how to program using Pico SDK, complete PDF based API documentation,
and for all the details and documentation of more advanced features.

See [Online Pico SDK API docs](todo) for HTML API documentation.

# Example code

See https://github.com/raspberrypi/pico-examples for example code you can build.

# Quick-start your own project

These instructions are exteremly terse, and Linux based only. For detailed steps,
instructions for other platforms, and just in general, we recommend you see [Getting Started with the Raspberry Pi Pico](rptl.io/picogs)

1. Install CMake (at least version 3.12), and GCC cross compiler
   ```
   sudo apt install cmake gcc-arm-none-eabi
   ```
1. Set up your project to point to this SDK
   
   * By cloning the SDK locally (most common)
      1. `git clone` the repository
      1. Copy [pico_sdk_import.cmake](https://github.com/raspberrypi/pico-sdk/blob/pre_release/external/pico_sdk_import.cmake)
         from the SDK into your project directory
      2. Set `PICO_SDK_PATH` to the SDK location in your environment, or pass it (`-DPICO_SDK_PATH=`) to cmake later.
      3. Setup a `CMakeLists.txt` like:

          ```cmake
          cmake_minimum_required(VERSION 3.12)

          # initialize the SDK based on PICO_SDK_PATH
          # note: this must happen before project()
          include(pico_sdk_import.cmake)

          project(my_project)

          # initialize the Pico SDK
          pico_sdk_init()

          # rest of your project

          ```

   * With Pico SDK as a submodule
      1. Clone the SDK as a submodule called `pico-sdk`
      1. Setup a `CMakeLists.txt` like:

          ```cmake
          cmake_minimum_required(VERSION 3.12)

          # initialize pico_sdk from submodule
          # note: this must happen before project()
          include(pico-sdk/pico_sdk_init.cmake)

          project(my_project)

          # initialize the Pico SDK
          pico_sdk_init()

          # rest of your project

          ```

   * With automatic download from github
      1. Copy [pico_sdk_import.cmake](https://github.com/raspberrypi/pico-sdk/blob/pre_release/external/pico_sdk_import.cmake)
         from the SDK into your project directory
      1. Setup a `CMakeLists.txt` like:

          ```cmake
          cmake_minimum_required(VERSION 3.12)

          # initialize pico_sdk from GIT
          # (note this can come from environment, CMake cache etc)
          set(PICO_SDK_FETCH_FROM_GIT on)

          # pico_sdk_import.cmake is a single file copied from this SDK
          # note: this must happen before project()
          include(pico_sdk_import.cmake)

          project(my_project)

          # initialize the Pico SDK
          pico_sdk_init()

          # rest of your project

          ```

3. Setup a CMake build directory.
      For example, if not using an IDE:
      ```
      $ mkdir build
      $ cd build
      $ cmake ..
      ```

4. Write your code (see https://github.com/raspberrypi/pico-examples or the [Pico C/C++ SDK](https://rptl.io/picoc) documentation)

   About the simplest you can do is a single source file (e.g. hello_world.c)

   ```c
   #include <stdio.h>
   #include "pico/stdlib.h"

   int main() {
       setup_default_uart();
       printf("Hello, world!\n");
       return 0;
   }
   ```
   And add the following to your `CMakeLists.txt`:

   ```cmake
   add_executable(hello_world
       hello_world.c
   )

   # Add pico_stdlib library which aggregates commonly used features
   target_link_libraries(hello_world pico_stdlib)

   # create map/bin/hex/uf2 file in addition to ELF.
   pico_add_extra_outputs(hello_world)

   ```



5. Make your target from the build directory you created.
      ```sh
      $ make hello_world
      ```

6. You now have `hello_world.elf` to load via a debugger, or `hello_world.uf2` that can be installed and
run on your Raspberry Pi Pico via drag and drop.